#!/usr/bin/python

import argparse
import os
import sys
import shutil


import pandas as pd
#import dask.dataframe as dd
from datetime import datetime, timedelta
from pathlib import Path

from unit_checker import UnitChecker
from func_tools import *
import dateutil.relativedelta

elements = ["CO", "SO2", "NO2", "PM2.5", "PM10", "O3"]

def split_date(df):
    """
    Function to split the date into separate Hour, Day, Month, Year columns
    """
    print(df)
    df["DATE"] = pd.to_datetime(df["DATE"], format="%Y%m%d%H")
    df['HOUR'] = df["DATE"].dt.hour.astype(str).str.zfill(2)
    df['DAY'] = df["DATE"].dt.day.astype(str).str.zfill(2)
    df['MONTH'] = df["DATE"].dt.month.astype(str).str.zfill(2)
    df['YEAR'] = df["DATE"].dt.year.astype(str).str.zfill(4)
    return df

# Second make a dataframe for all the element for the whole year
def process_dataframe(all_files):
    """
    Open all files and create a single dataframe per year with the structurs
        
        DATE | CONCENTRATION | PARAMETER | STATION
          -  |       -       |     -     |    -
    """
    dfs = []
    for file in all_files:
        filename = str(file).split("/")[-1]
        element = str(file).split("/")[-2] # Get the element
        station = filename.split(".")[0]

        df = pd.read_csv(file, sep=" ", names=["DATE", "CONCENTRATION"], index_col=False)
        df["PARAMETER"] = element
        df["STATION"] = station
        dfs.append(df)


    df = pd.concat(dfs, axis=0)
    return df

# Third add the station info to the dataframe
def adding_station_info(df, station_dir, element, data_type="assimilation"):
    """
    Add station information to a dataframe e.g. lat, lon, alt values
    """
    for station_file in os.listdir(station_dir):
        if element == "PM2.5":
            station_element = "PM25"
        else: 
            station_element = element
        if station_file.startswith('stations.{}.background_{}_set'.format(str(station_element), data_type)):
            station_filename = station_dir.joinpath(station_file)
            print("Station File being used {}".format(station_filename))
    station_df = pd.read_csv(station_filename, sep = ' ', names=["STATION", "LON", "LAT", "ALT(m)", "Background", "Info"])
    print(station_df)
    df = station_df.merge(df, how="right")
    print(df)
    df = split_date(df)
    df = df.drop_duplicates()
    print("After")
    print(df)
    return df

# This is run first
def formatting_yearly_vradata(in_dir, date, data_type):
    """
    The function that formats the data

    * First we find all files for that particular year
    * Then pass that through processing datafame
    * Before adding station info
    """
    print("Using station list for year {}".format(date))
    station_dir = in_dir.joinpath("stations_{}".format(date))
    #in_dir = in_dir.joinpath(data_type)
    all_files = []

    year_dir = in_dir.joinpath(str(date))
    
    element_dfs = []
    for element in elements:
        # Get all files for each element
        if element == "PM2.5":
            station_element = "PM25"
        else: 
            station_element = element
        print(element)
        element_dir = year_dir.joinpath(station_element)
        print(element_dir)
        for path, subdirs, files in os.walk(element_dir):
            for name in files:
                if str(date) in name:
                    filename = Path(path).joinpath(name)
                    all_files.append(filename)
        print(all_files)
        # Need to add an if CO then convert from milligrams to micrograms
        # Also add the unit checker that I did for IRA too
        # But I will wait for the nord3 run first
        ## Format Section
        df = process_dataframe(all_files=all_files)
        df = adding_station_info(df, station_dir, element)
        df = df.dropna(subset=["LON", "LAT", "ALT(m)", "Background", "Info"])
        element_dfs.append(df)
    final_df = pd.concat(element_dfs, axis=0)

    return final_df

def filtering(in_dir, date, df):
    """
    Applying filtering and checker steps to the data
    """
    # Remove missing value marker
    df = df[df["CONCENTRATION"] != -100]
    # Read in all data for the previous year
    base_date = datetime.strptime(str(date), "%Y")
    startdate = base_date - timedelta(days=365) 
    enddate = base_date
    startdate = startdate.strftime("%Y%m%d")
    enddate = enddate.strftime("%Y%m%d")
    dates = pd.date_range(startdate, enddate)
    years = list(dict.fromkeys(dates.year))
    dfs = []
    for year in years:
        print(date)
        month_df = formatting_yearly_vradata(in_dir, year, "assimilation")
        dfs.append(month_df)

    dfs.append(df)

    df_month = pd.concat(dfs)
    df_month.reset_index(inplace=True, drop=True)
    

    check = UnitChecker(df, df_month, date, "vra")
    check.unit_checker()

    df_changed_data = check.df_changed_values

    logged_dir = format_dir.joinpath("logged_data") # For later
    ## Assimilation section
    print("Formatting Assimilation Stations")

    if not os.path.exists(logged_dir):
        os.makedirs(logged_dir)
        os.system('chmod 765 -R {}'.format(logged_dir))

    logged_file = logged_dir.joinpath("{}_{}_updated_data.csv".format("vra", date))
    df_changed_data.to_csv(logged_file, sep=";", index=False)
    print(check.df)
    return check.df

def filter_statistics_provided_by_ineris(df, date, in_dir):
    """
    A functino to filter the outliers and underliers data supplied by ineris
    """
    year_dir = in_dir.joinpath(str(date))
    st_dir = year_dir.joinpath("statistical_outliers")
    print(df)
    # Search for both the element and outlier/underlier
    dfs = []
    
    elements = ["CO", "SO2", "NO2", "PM25", "PM10", "O3"]
    for statistic in ["outlier", "underlier"]:
        for e in elements:
            print(e)
            print(statistic)
            try:
                st_df = pd.read_csv(st_dir.joinpath("pol_{}.{}".format(e, statistic)), names = ["dataset", "value", "limit"], delim_whitespace=True, index_col=False)
            except FileNotFoundError:
                print("File doesn't exist for element {} and statistic {}".format(e, statistic))
                continue
            data_dict = {"STATION":[], "DATE":[], "PARAMETER":[]}
            for i in st_df["dataset"]:
                data = i.split("/")[-1]
                station = data.split(".")[0]
                date = data.split(":")[-1]
                # Append outliers to new df
                date = datetime.strptime(date,"%Y%m%d%H").strftime("%Y-%m-%d %H:%M:%S")
                data_dict["STATION"].append(station)
                data_dict["DATE"].append(date)
                data_dict["PARAMETER"].append(e)
            a_df = pd.DataFrame(data_dict)
            
            dfs.append(a_df)
    stats_df = pd.concat(dfs)
    print(df)
    stats_df["keep"] = 0
    stats_df["DATE"] = pd.to_datetime(stats_df["DATE"], format="%Y-%m-%d %H:%M:%S")
    df = df.merge(stats_df, on=["STATION", "DATE", "PARAMETER"], how="left")
    df["keep"] = df["keep"].fillna(1)
    fdf = df[df["keep"] == 1]
    odf = df[df["keep"] == 0]
    odf.to_csv("/gpfs/scratch/bsc32/bsc032107/vra-filtered/daily/removed_data.csv", sep=';', quotechar='"', encoding='latin-1', index=False, header=True)
    print(fdf)
    return fdf

# Finally output all days
def output_date_csvs(df, format_dir, dates):
    """
    Output each day as a separate file
    """

    #use to_period
    per = df['DATE'].dt.to_period("d")
    print(per)
    #group by that period
    agg = df.groupby([per])
    for day, group in agg:
        #this simple save the data
        datep = day[0].strftime("%Y-%m-%d")#.replace('-', '')
        if datep in dates:
            print(day)
            print("Saving to directory")
            datep = datep.replace('-', '')
            output_file = format_dir.joinpath("obscamsvra_{}.csv".format(datep))
            group = group.sort_values(by=["PARAMETER", "DATE",  "STATION"])
            group = group.drop_duplicates()
            group.to_csv(output_file, sep=';', quotechar='"', encoding='latin-1', index=False, header=True)


def getOptions(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-s", "--startdate", help="Start date for formatting")
    parser.add_argument("-e", "--enddate", help="End date for formatting")
    parser.add_argument("-v", "--validated", help="Whether we want to process the validated data")
    options = parser.parse_args(args)
    return options

if __name__ == "__main__":
    print("Arguments to be parsed")
    print(sys.argv)

    options = getOptions(sys.argv[1:])
    startdate = options.startdate
    enddate = options.enddate
    valid = options.validated

    dates= pd.date_range(startdate, enddate)
    # Only get the years
    years = list(dict.fromkeys(dates.year))


    in_dir = Path("/gpfs/scratch/bsc32/bsc032107/vra/original_files/")
    format_dir = Path('/gpfs/scratch/bsc32/bsc032107/vra-filtered/daily/')

    # Create format dir if it doesn't already exist
    
    if not os.path.exists(format_dir):
        os.makedirs(format_dir)
        os.system('chmod 765 -R {}'.format(format_dir))
    data_type = "assimilation"

    for year in years:
        # Format yearly files separately
        print(year)
        df = formatting_yearly_vradata(in_dir, year, data_type)
        df = filtering(in_dir, year, df)
        df = filter_statistics_provided_by_ineris(df, year, in_dir)
        output_date_csvs(df, format_dir, dates)
