"""
Python class for Checking Eionet data required for preprocessor
"""
from collections.abc import MutableMapping
from datetime import datetime, timedelta

import numpy as np
import pandas as pd

import dateutil.relativedelta
import math





# empty slot marker
_EMPTY = object()


class RangeMap(MutableMapping):
    """
    Map points to values, and find values for points in O(1) constant time

    The map requires a fixed minimum lower and maximum upper bound for
    the ranges. Range boundaries are quantized to a fixed step size. Gaps
    are permitted, when setting overlapping ranges last range set wins.

    """
    def __init__(self, map=None, lower=-5000000.0, upper=5000000.0, step=0.05):
        self._mag = 10 ** -round(math.log10(step) - 1)  # shift to integers
        self._lower, self._upper = round(lower * self._mag), round(upper * self._mag)
        self._step = round(step * self._mag)
        self._steps = (self._upper - self._lower) // self._step
        self._table = [_EMPTY] * self._steps
        self._len = 0
        if map is not None:
            self.update(map)

    def __len__(self):
        return self._len

    def _map_range(self, r):
        low, high = r
        start = round(low * self._mag) // self._step
        stop = round(high * self._mag) // self._step
        if not self._lower <= start < stop <= self._upper:
            raise IndexError('Range outside of map boundaries')
        return range(start - self._lower, stop - self._lower)

    def __setitem__(self, r, value):
        for i in self._map_range(r):
            self._len += int(self._table[i] is _EMPTY)
            self._table[i] = value

    def __delitem__(self, r):
        for i in self._map_range(r):
            self._len -= int(self._table[i] is not _EMPTY)
            self._table[i] = _EMPTY

    def _point_to_index(self, point):
        point = round(point * self._mag)
        if not self._lower <= point <= self._upper:
            raise IndexError('Point outside of map boundaries')
        return (point - self._lower) // self._step

    def __getitem__(self, point_or_range):
        if isinstance(point_or_range, tuple):
            low, high = point_or_range
            r = self._map_range(point_or_range)
            # all points in the range must point to the same value
            value = self._table[r[0]]
            if value is _EMPTY or any(self._table[i] != value for i in r):
                raise IndexError('Not a range for a single value')
        else:
            value = self._table[self._point_to_index(point_or_range)]
            if value is _EMPTY:
                raise IndexError('Point not in map')
        return value

    def __iter__(self):
        low = None
        value = _EMPTY
        for i, v in enumerate(self._table):
            pos = (self._lower + (i * self._step)) / self._mag
            if v is _EMPTY:
                if low is not None:
                    yield (low, pos)
                    low = None
            elif v != value:
                if low is not None:
                    yield (low, pos)
                low = pos
                value = v
        if low is not None:
            yield (low, self._upper / self._mag)




class UnitChecker:
    """
    An object to chek eionet data to see if we need to update the data
    """
    def __init__(self, df, df_month, date, data_type):
        self.df = df.reset_index(drop=True)
        self.df_month = df_month
        self.df_changed_values = None
        self.date = date
        self.data_type = data_type

    def unit_checker(self):
        """
        A function that checks the units of a station and adjusts them if it believes they are wrong

        Parameters
        -----------
        df - original data in dataframe format with all elements

        Outputs
        -----------
        1. Change of concentration on original df 
        2. Dataframe with all changes added by checker
        """

        # If the median is above a certain threshold divide all values by 1000
        # If a vslue is a certain amount above the median threshold than apply some calculation

        # Update the CO from micrograms to milligrams
        
        
        
        self.df["OLD_CONCENTRATION"] = self.df["CONCENTRATION"]
        self.df["CONCENTRATION"] = self.df["CONCENTRATION"].where(self.df["PARAMETER"]!="CO", self.df["CONCENTRATION"] / 1000)

	
        grouped_month = self.df_month.groupby(["PARAMETER", "STATION"]).median(numeric_only=True)

        # Monthly median per paramter
        param_medians = {
            "CO": self.df_month[self.df_month["PARAMETER"] == "CO"]["CONCENTRATION"].median(),
            "NO2": self.df_month[self.df_month["PARAMETER"] == "NO"]["CONCENTRATION"].median(),
            "O3": self.df_month[self.df_month["PARAMETER"] == "O3"]["CONCENTRATION"].median(),
            "SO2": self.df_month[self.df_month["PARAMETER"] == "SO2"]["CONCENTRATION"].median(),
            "PM10": self.df_month[self.df_month["PARAMETER"] == "PM10"]["CONCENTRATION"].median(),
            "PM25": self.df_month[self.df_month["PARAMETER"] == "PM25"]["CONCENTRATION"].median(),
        }

        # Construct the range map, so if the diff is between these values return the scale factor
        #d = RangeMap({
        #    (-500, 500): 1,
        #    (500, 5000): 0.001,
        #    (5000, 50000): 0.0001,
        #})
        

        # Check to see if value is radically different to monthly median
        median_anomolies = np.ones(len(self.df))
        for row in range(len(self.df)):
            station = self.df.loc[row, "STATION"]
            param = self.df.loc[row, "PARAMETER"]

            conc = grouped_month.loc[(param, station), :]["CONCENTRATION"]
            

            if (5000 > conc / param_medians[param] > 500) & (5000 > self.df.loc[row, "CONCENTRATION"] / param_medians[param] > 500):
                median_anomolies[row] = 0.001
            elif (50000 > conc / param_medians[param] > 5000) & (50000 > self.df.loc[row, "CONCENTRATION"] / param_medians[param] > 5000):
                median_anomolies[row] = 0.0001
            elif (5000 > param_medians[param] / conc > 500) & (5000 > param_medians[param] / self.df.loc[row, "CONCENTRATION"] > 500):
                median_anomolies[row] = 1000
            elif (50000 > param_medians[param] / conc > 5000) & (50000 > param_medians[param] / self.df.loc[row, "CONCENTRATION"] > 5000):
                median_anomolies[row] = 10000
                
                


        self.df["Multiplying Concentration"] = median_anomolies

        
        self.df["CONCENTRATION"] = self.df["CONCENTRATION"] * self.df["Multiplying Concentration"]
        self.df["NEW_CONCENTRATION"] = self.df["CONCENTRATION"]


        self.df_changed_values = self.df[self.df["OLD_CONCENTRATION"] != self.df["NEW_CONCENTRATION"]]
        columns_of_interest = ["STATION", "DATE", "PARAMETER", "OLD_CONCENTRATION", "NEW_CONCENTRATION"]
        self.df_changed_values = self.df_changed_values[columns_of_interest]

        dropping_columns = ["Multiplying Concentration", "OLD_CONCENTRATION", "NEW_CONCENTRATION"]
        self.df = self.df.drop(dropping_columns, axis=1)

        return None
