#!/usr/bin/python

import os
import os.path
import argparse
import sys
import pandas as pd
import numpy as np
import xarray as xr
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
from glob import glob
from nes import create_nes
from pathlib import Path

AC_CONVENTIONS = {
    "O3": "sconco3",
    "NO2": "sconcno2",
    "CO": "sconcco",
    "SO2": "sconcso2",
    "PM10": "pm10",
    "PM25": "pm2p5"
}



class MeteoFranceFormat:

    def __init__(self, in_dir, out_dir, month):
        self.in_dir = in_dir
        self.out_dir = out_dir
        self.month = month
        self.units = "ug m-3"
        date_indatetime = datetime.strptime(self.month, "%Y%m")
        self.next_month = (date_indatetime + relativedelta(months = 1)).strftime("%Y%m")

    def read_csv(self):
        """
        Read dataframe
        """ 
        start_date = (datetime.strptime("{}01".format(month), "%Y%m%d") - timedelta(days=1)).strftime("%Y%m%d")

        data_files = []
        for filename in os.listdir(self.in_dir):
            if filename.startswith('obscamsvra_{}'.format(self.month)):
                print(filename)
                data_files.append(self.in_dir.joinpath(filename))

        self.df = pd.concat((
            pd.read_csv(filename, delimiter=';', 
            dtype={
                'YEAR':str, 
                "MONTH":str, 
                'DAY':str, 
                'HOUR':str
                }) for filename in data_files))
        self.elements = list(self.df["PARAMETER"].unique())
        self.df.reset_index(drop=True, inplace=True)
    
    def preprocess_dataframe(self):
        """
        Preprocess dataframe 
        """
        self.df['time'] = self.df[['YEAR', 'MONTH', 'DAY', 'HOUR']].agg(''.join, axis=1)
        self.df["time"] = pd.to_datetime(self.df["time"], format="%Y%m%d%H")

        self.df_new = self.df.copy()
        self.df_new.drop(columns=["YEAR","MONTH","DAY","HOUR"], inplace=True)
        self.df_new.rename(columns={"CONCENTRATION":"CONC"}, inplace=True)
        self.df_new = self.df_new.sort_values(by=["time", "STATION"])
        #Remove duplicate records
        self.df_new.drop_duplicates(subset=["time", "PARAMETER", "CONC"], inplace=True)
        self.df_new = self.df_new.groupby(["time","STATION", "PARAMETER"]).first().reset_index()
        self.df_new.set_index(["time", "PARAMETER"], inplace=True)


    def process_station_info(self):
        self.df_location = self.df.loc[:, ["STATION", "LAT", "LON", "ALT(m)", "PARAMETER"]]
        self.df_location["STATION"] = self.df_location["STATION"].astype(str)
        # Take the first value will update to metadata 
        self.df_location = self.df_location.groupby(["STATION", "PARAMETER"]).first().reset_index()

        self.df_location.drop_duplicates(inplace=True)
        self.df_location.set_index("STATION", inplace=True)
        print(self.df_location.reset_index()["STATION"].value_counts())

    @staticmethod
    def create_folders(directory):
        if not os.path.exists(directory):
            os.makedirs(directory)
            os.system('chmod 755 -R {}'.format(directory))


    def output_netcdf(self, element):
        """
        Create nes object and output to netcdf
        """
        # Rename the element to be more inline with the AC conventions
        element_ac = AC_CONVENTIONS[element]

        element_dir = self.out_dir.joinpath(element_ac)
        self.create_folders(element_dir)
        outfile = element_dir.joinpath('{}_{}.nc'.format(element_ac, self.month))
        element_df = self.df_new.xs(element, level="PARAMETER").pivot(columns="STATION", values="CONC")
        # Convert CO to micrograms
        if element == "CO":
            element_df = element_df*1000
        #Reindex the df to have all the correct dates
        element_df = element_df.reindex(pd.date_range('{}01'.format(self.month),'{}01'.format(self.next_month), freq='H', closed="left"))

        element_location_df = self.df_location[self.df_location['PARAMETER'] == element]
        lat = element_location_df.loc[element_df.columns, :]["LAT"]
        lon = element_location_df.loc[element_df.columns, :]['LON']
        print(element_location_df.reset_index()["STATION"].value_counts())

        element_data = create_nes(comm=None, info=False, projection=None, parallel_method='X',
            lat=lat, lon=lon, times=element_df.index)


        metadata = {'station_name': {'data': element_location_df.loc[element_df.columns, :].index,
                             'dimensions': ('station',), 
                             'dtype': str},
                'alt': {'data': element_location_df.loc[element_df.columns, "ALT(m)"],
                             'dimensions': ('station',),
                             'units': 'meters',
                             'standard_name': 'altitude',
                             'dtype':int},
                element_ac: {'data': element_df.loc[:, :].to_numpy(),
                                    'dimensions': ('time', 'station',), 
                                    'units':self.units,
                                    'dtype': float}}
    
        element_data.variables = metadata
        
        element_data.to_netcdf(outfile, info=True)

def getOptions(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-d", "--date", type=str, help="Date that corresponds to the month and previous month you want procesed.")
    parser.add_argument("-i", "--indir", type=str, help="In directory - e.g. original files.")
    parser.add_argument("-o", "--outdir", type=str, help="Out directory hourly files.") 
    parser.add_argument("-n", "--noprevmonths", type=int, help="Number of previous months to be processed")
    options = parser.parse_args(args)
    return options


if __name__ == "__main__":
    import sys
    options = getOptions()
    in_dir = Path(options.indir) #Path('/esarchive/obs/meteofrance/eionet-cams2_40/original_files/')
    out_dir = Path(options.outdir) #Path('/esarchive/obs/meteofrance/eionet-cams2_40/hourly/')
    date = options.date #"20221124"
    number_of_previous_months = options.noprevmonths #1
    # Get only the month for this month
    this_month = date[:-2]
    print(this_month)

    # Create a list with this month
    months = [this_month]
    # And append previous months to be processed to that list 
    for n in range(0, number_of_previous_months):
        # Get the previous month to preprocess that too
        date_indatetime = datetime.strptime(this_month, "%Y%m")
        previous_month = (date_indatetime - relativedelta(months = n+1)).strftime("%Y%m")
        print(previous_month)
        months.append(previous_month)

    # Run through the months a process them 
    for month in months:
        # Process for this month
        print("Processing for month {}".format(month))
        meteo = MeteoFranceFormat(in_dir, out_dir, month)
        print("Read all daily files")
        meteo.read_csv()
        print("Process the dataframe")
        meteo.preprocess_dataframe()
        print("Get the station information")
        meteo.process_station_info()
        for pollutant in meteo.elements:
            print("Outputting netcdf for pollutant {}".format(pollutant))
            meteo.output_netcdf(pollutant)
